import matplotlib.pyplot as plt
import argparse
import warnings
from orcanet.history import HistoryHandler
from orcanet.utilities.visualization import TrainValPlotter


class Summarizer:
    """
    Summarize one or more trainings by giving their orcanet folder(s).

    - Plot the training and validation curves in a single plot and show them
    - Print info about the best and worst epochs

    Attributes
    ----------
    folders : str or List, optional
        Path to a orcanet folder, or to multiple folder as a list.
        Default: CWD.
    metric : str
        The metric to plot. Default: loss.
    smooth : int, optional
        Apply gaussian blur to the train curve with given sigma.
    labels : List, optional
        Labels for each folder.
    noplot : bool
        Dont plot the train/val curves [default: False].
    width : float
        Scaling of the width of the curves and the marker size [default: 1].

    """
    def __init__(self, folders,
                 metric="loss",
                 smooth=None,
                 labels=None,
                 noplot=False,
                 width=1.):
        self.folders = folders
        self.metric = metric
        self.smooth = smooth
        self.labels = labels
        self.noplot = noplot
        self.width = width
        self._tvp = None

    def summarize(self):
        if not self.noplot:
            self._tvp = TrainValPlotter()

        min_stats, max_stats = [], []
        print("Reading stats of {} trainings...".format(len(self._folders)))
        for folder_no in range(len(self._folders)):
            try:
                min_stat, max_stat = self._summarize_folder(folder_no)
                min_stats.append(min_stat)
                max_stats.append(max_stat)
            except OSError:
                warnings.warn("Can not summarize {}, skipping..."
                              "".format(self._folders[folder_no]))

        min_stats.sort()
        print("\nMinimum\n-------")
        print("{}   \t{}\t{}\t{}".format(" ", "Epoch", self._full_metric, "name"))
        for i, stat in enumerate(min_stats, 1):
            print("{} | \t{}\t{}\t{}".format(i, stat[2], stat[0], stat[1]))

        max_stats.sort(reverse=True)
        print("\nMaximum\n-------")
        print("{}   \t{}\t{}\t{}".format(" ", "Epoch", self._full_metric, "name"))
        for i, stat in enumerate(max_stats, 1):
            print("{} | \t{}\t{}\t{}".format(i, stat[2], stat[0], stat[1]))

        if not self.noplot:
            self._tvp.apply_layout(x_label="Epoch",
                                   y_label=self._metric_name,
                                   grid=True,
                                   legend=True)
            plt.show()

    @property
    def _metric_name(self):
        """ E.g. loss """
        if self.metric.startswith("train_"):
            metric = self.metric[6:]
        elif self.metric.startswith("val_"):
            metric = self.metric[4:]
        else:
            metric = self.metric
        return metric

    @property
    def _full_metric(self):
        """ E.g. val_loss """
        if not (self.metric.startswith("train_") or
                self.metric.startswith("val_")):
            full_metric = "val_" + self.metric
        else:
            full_metric = self.metric
        return full_metric

    @property
    def _folders(self):
        """ Get a list of folders. """
        if not self.folders:
            folders = "./"
        else:
            folders = self.folders

        if isinstance(folders, str):
            folders = [folders]
        return folders

    @property
    def _labels(self):
        """ Get a list of labels. """
        if self.labels is None:
            return self._folders
        else:
            return self.labels

    def _summarize_folder(self, folder_no):
        label = self._labels[folder_no]
        folder = self._folders[folder_no]
        if len(self._labels) == 1:
            train_label, val_label = "training", "validation"
        else:
            train_label, val_label = None, label

        hist = HistoryHandler(folder)
        summary_data = hist.get_summary_data()
        full_train_data = hist.get_train_data()

        train_data = [full_train_data["Batch_float"],
                      full_train_data[self._metric_name]]
        val_data = [summary_data["Epoch"],
                    summary_data[self._full_metric]]

        smry_met_name = self._full_metric
        max_line = hist.get_best_epoch_info(metric=smry_met_name,
                                            mini=False)
        min_line = hist.get_best_epoch_info(metric=smry_met_name, mini=True)

        min_stat = [min_line[smry_met_name], label, min_line["Epoch"]]
        max_stat = [max_line[smry_met_name], label, max_line["Epoch"]]

        if not self.noplot:
            self._tvp.plot_curves(train_data=train_data,
                                  val_data=val_data,
                                  train_label=train_label,
                                  val_label=val_label,
                                  smooth_sigma=self.smooth,
                                  tlw=0.5*self.width,
                                  vlw=0.5*self.width,
                                  vms=3*self.width**0.5)
        return min_stat, max_stat

    def summarize_dirs(self):
        """
        Get the best and worst epochs of all given folders as a dict.

        Returns
        -------
        minima : dict
            Keys : Name of folder.
            Values : [Epoch, metric] of where the metric is lowest.
        maxima : dict
            As above, but for where the metric is highest.

        """
        minima, maxima = {}, {}
        for folder in self._folders:
            hist = HistoryHandler(folder)
            smry_met_name = self._full_metric
            try:
                max_line = hist.get_best_epoch_info(metric=smry_met_name,
                                                    mini=False)
                min_line = hist.get_best_epoch_info(metric=smry_met_name,
                                                    mini=True)
            except OSError as e:
                warnings.warn(str(e))
                continue

            minima[folder] = [min_line["Epoch"], min_line[smry_met_name]]
            maxima[folder] = [max_line["Epoch"], max_line[smry_met_name]]
        return minima, maxima


def main():
    parser = argparse.ArgumentParser(
        description=str(Summarizer.__doc__),
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('folders', type=str, nargs='*')
    parser.add_argument('-metric', type=str, nargs="?")
    parser.add_argument('-smooth', nargs="?", type=int)
    parser.add_argument('-width', nargs="?", type=float)
    parser.add_argument('-labels', nargs="*", type=str)
    parser.add_argument('-noplot', action="store_true")
    args = vars(parser.parse_args())
    for key in list(args.keys()):
        if args[key] is None:
            args.pop(key)

    Summarizer(**args).summarize()


if __name__ == '__main__':
    main()



import keras as ks
import keras.layers as layers
import keras.backend as K
import inspect

import orcanet.builder_util.layer_blocks as layer_blocks


class BlockBuilder:
    """
    Can build single-input block-wise sequential neural network.

    The body section of the network always ends with a flatten layer.

    Attributes
    ----------
    body_defaults : dict or None
        Default values for all blocks in the body section of the model.
    head_defaults : dict or None
        Default values for all blocks in the head section of the model.

    """

    def __init__(self, body_defaults=None, head_defaults=None):
        """ Set dicts with default values for the layers of the model.
        """
        self.all_blocks = {
            "conv_block": layer_blocks.ConvBlock,
            "dense_block": layer_blocks.DenseBlock,
            "resnet_block": layer_blocks.ResnetBlock,
            "resnet_bneck_block": layer_blocks.ResnetBnetBlock,
        }

        self._check_arguments(body_defaults)
        self._check_arguments(head_defaults)

        self.body_defaults = body_defaults
        self.head_defaults = head_defaults

    def build(self, input_shape, body_configs, head_arch, head_arch_args=None):
        """
        Build the whole model, using the default values when arguments
        are missing in the layer_configs.

        Parameters
        ----------
        input_shape : dict
            Name and shape of the input layer.
        body_configs : list
            List of configurations for the blocks in the body of the model.
            Each element in the list is a dict and will result in one block
            connected to the previous one. The dict has to contain the type
            of the block, as well as any arguments required by that
            specific block type.
        head_arch : str, optional
            Specifies the architecture of the head. None for no head.
        head_arch_args : dict or None
            Required arguments for the given head_arch.

        Returns
        -------
        model : keras model

        """
        if not isinstance(input_shape, dict):
            raise TypeError(
                "input_shapes must be a dict, not ", type(input_shape))
        if not len(input_shape) == 1:
            raise TypeError(
                "input_shapes must have length 1, not ", len(input_shape))

        input_layer = layer_blocks.input_block(input_shape)[0]

        x = input_layer
        for layer_config in body_configs:
            x = self.attach_block(x, layer_config)

        model_body = ks.models.Model(input_layer, x)
        last_b_layer = model_body.layers[-1]
        last_b_layer.name = "BODYEND_" + last_b_layer.name

        if head_arch is None:
            outputs = x

        else:
            if head_arch_args is None:
                head_arch_args = {}
            outputs = self.attach_output_layers(x,
                                                head_arch, **head_arch_args)

        model = ks.models.Model(inputs=input_layer, outputs=outputs)
        return model

    def attach_block(self, layer, layer_config, is_output=False):
        """
        Attach a block to the given layer based on the layer config.

        Will use the default values given during initialization if they are not
        present in the layer config.

        Parameters
        ----------
        layer : keras layer
            Layer to attach the block to.
        layer_config : dict
            Configuration of the block to attach. The dict has to contain
            the type of the block, as well as any arguments required by that
            specific block.
        is_output : bool
            Whether to use the default values for the head layers or
            the body layers.

        Returns
        -------
        x : keras layer

        """
        if is_output:
            defaults = self.head_defaults
        else:
            defaults = self.body_defaults
        filled = self._with_defaults(layer_config, defaults)
        block = self._get_blocks(filled.pop("type"))
        x = block(**filled)(layer)
        return x

    def attach_output_layers(self, layer, head_arch, **head_arch_args):
        """
        Append the head dense layers to the network.

        The output arhcitectures are listed below.

        Parameters
        ----------
        layer : keras layer
            The Keras layer instance to attach the head to.
        head_arch : str
            Specifies the architecture of the head layers.
        head_arch_args : dict
            Arguments for the given output_arch. Given as the output_args
            dict in the model toml file.

        Returns
        -------
        outputs : list
            List of the keras output layers of the network.

        """
        if head_arch == "categorical":
            # categorical problem
            assert head_arch_args is not None, "No output_kwargs given"
            outputs = self.attach_output_cat(layer, **head_arch_args)

        elif head_arch == "gpool" or head_arch == "gpool_categ":
            outputs = self.attach_output_gpool_categ(layer, **head_arch_args)

        elif head_arch == "gpool_reg":
            outputs = self.attach_output_gpool_reg(layer, **head_arch_args)

        elif head_arch == "regression_error":
            # regression with error estimation, two outputs for each label
            assert head_arch_args is not None, "No output_kwargs given"
            outputs = self.attach_output_reg_err(layer, **head_arch_args)

        else:
            raise ValueError("Unknown head_arch: " + str(head_arch))

        return outputs

    def attach_output_cat(self, layer, categories, output_name,
                          flatten=True):
        """ Small dense network for multiple categories. """
        if flatten:
            x = layers.Flatten()(layer)
        else:
            x = layer

        x = self.attach_block(
            x, {"type": "dense_block", "units": 128},
            is_output=True)

        x = self.attach_block(
            x, {"type": "dense_block", "units": 32, "dropout": None},
            is_output=True)

        out = layers.Dense(
            units=categories,
            activation='softmax',
            kernel_initializer='he_normal',
            name=output_name)(x)
        return [out, ]

    @staticmethod
    def attach_output_gpool_categ(layer, categories, output_name, dropout=None):
        """ Global Pooling + 1 dense layer (like in resnet). """
        x = layers.GlobalAveragePooling2D()(layer)
        if dropout is not None:
            x = layers.Dropout(dropout)(x)

        out = layers.Dense(
            units=categories,
            activation='softmax',
            kernel_initializer='he_normal',
            name=output_name)(x)
        return [out, ]

    def attach_output_gpool_reg(
            self, layer, output_name, output_neurons=1, unit_list=None):
        """
        Global Pooling, followed by dense layer(s) for regression.

        Parameters
        ----------
        layer
        output_name : str
            Name that will be given to the output layer of the network.
        output_neurons : int
            Number of neurons in the last layer. Default: 1.
        unit_list : List, optional
            A list of ints. Add additional Dense layers after the gpool
            with this many units in them. E.g., [64, 32] would add
            two Dense layers, the first with 64 neurons, the secound with
            32 neurons.

        """
        x = layers.GlobalAveragePooling2D()(layer)

        if unit_list is not None:
            for units in unit_list:
                x = self.attach_block(
                    x, {"type": "dense_block", "units": units},
                    is_output=True)

        out = layers.Dense(units=output_neurons, name=output_name)(x)

        return [out, ]

    def attach_output_reg_err(self, layer, output_names, flatten=True):
        """ Double network for regression + error estimation. """
        if flatten:
            flatten = layers.Flatten()(layer)
        else:
            flatten = layer
        outputs = []

        # Network for the labels
        x = self.attach_block(
            flatten, {"type": "dense_block", "units": 128},
            is_output=True)

        x = self.attach_block(
            x, {"type": "dense_block", "units": 32, "dropout": None},
            is_output=True)

        for name in output_names:
            output_label = layers.Dense(units=1, name=name)(x)
            outputs.append(output_label)

        # Network for the errors of the labels
        x_err = layers.Lambda(lambda a: K.stop_gradient(a))(flatten)

        x_err = self.attach_block(
            x_err, {"type": "dense_block", "units": 128},
            is_output=True)

        x_err = self.attach_block(
            x_err, {"type": "dense_block", "units": 64},
            is_output=True)

        x_err = self.attach_block(
            x_err, {"type": "dense_block", "units": 32, "dropout": None},
            is_output=True)

        for i, name in enumerate(output_names):
            output_label_error = layers.Dense(
                units=1,
                activation='linear',
                name=name + '_err_temp')(x_err)
            # Predicted label gets concatenated with its error (needed for loss function)
            output_label_merged = layers.Concatenate(name=name + '_err')(
                [outputs[i], output_label_error])
            outputs.append(output_label_merged)

        return outputs

    def _with_defaults(self, config, defaults):
        """ Make a copy of a layer config and complete it with default values
        for its block, if they are missing in the layer config.
        """
        conf = dict(config)

        if config is not None and "type" in config:
            block_name = config["type"]
        elif defaults is not None and "type" in defaults:
            block_name = defaults["type"]
            conf["type"] = defaults["type"]
        else:
            raise KeyError("No layer block type specified")

        block = self._get_blocks(block_name)
        args = inspect.getfullargspec(block).args
        if "self" in args:
            args.remove("self")

        if defaults is not None:
            for key, val in defaults.items():
                if key in args and key not in conf:
                    conf[key] = val

        return conf

    def _get_blocks(self, name=None):
        """ Get the block class/function depending on the name. """
        if name is None:
            return self.all_blocks
        elif name in self.all_blocks:
            return self.all_blocks[name]
        else:
            raise NameError("Unknown block type: {}, must be one of {}"
                            "".format(name, self.all_blocks.keys()))

    def _check_arguments(self, defaults):
        """ Check if given defaults appear in at least one block. """
        if defaults is None:
            return
        # possible arguments for all blocks
        psb_args = ["type", ]
        for block in self._get_blocks().values():
            args = inspect.getfullargspec(block).args
            for arg in args:
                if arg not in psb_args and arg != "self":
                    psb_args.append(arg)

        for t_def in defaults.keys():
            if t_def not in psb_args:
                raise NameError("Unknown argument: " + str(t_def))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import os
from unittest import TestCase
from unittest.mock import MagicMock
from keras.models import Model
import keras.layers as layers

from orcanet.core import Organizer
from orcanet.model_builder import ModelBuilder
from orcanet_contrib.custom_objects import get_custom_objects


class TestModel(TestCase):
    def setUp(self):
        """
        Make a .temp directory in the current working directory, generate
        dummy data in it and set up the cfg object.

        """
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.model_file = os.path.join(self.data_folder, "model_builder_test_CNN_cat.toml")

        self.input_shapes = {
            "input_A": (3, 3, 3, 3),
        }

        orga = Organizer(".")
        orga.io.get_input_shapes = MagicMock(return_value=self.input_shapes)
        orga.cfg.custom_objects = get_custom_objects()

        self.orga = orga

    def test_model_setup_CNN_model(self):
        orga = self.orga

        builder = ModelBuilder(self.model_file)
        model = builder.build(orga)

        self.assertEqual(model.input_shape[1:], self.input_shapes["input_A"])
        self.assertEqual(model.output_shape[1:], (2, ))
        self.assertEqual(len(model.layers), 14)
        self.assertEqual(model.optimizer.epsilon, 0.2)

    def test_merge_models(self):
        def build_model(inp_layer_name, inp_shape):
            inp = layers.Input(inp_shape, name=inp_layer_name)
            x = layers.Convolution3D(3, 3)(inp)
            x = layers.Flatten()(x)
            out = layers.Dense(1, name="out_0")(x)

            model = Model(inp, out)
            return model

        model_file = os.path.join(self.data_folder, self.model_file)
        builder = ModelBuilder(model_file)
        model1 = build_model("inp_A", self.input_shapes["input_A"])
        model2 = build_model("inp_B", self.input_shapes["input_A"])
        merged_model = builder.merge_models([model1, model2])

        for layer in model1.layers + model2.layers:
            if isinstance(layer, layers.Dense):
                continue
            merged_layer = merged_model.get_layer(layer.name)
            for i in range(len(layer.get_weights())):
                self.assertTrue(np.array_equal(layer.get_weights()[i],
                                               merged_layer.get_weights()[i]))



"""
For plotting keras and model related stuff.
"""

import numpy as np
import matplotlib.pyplot as plt

from orcanet.utilities.nn_utilities import get_layer_output


def plot_activations(model, samples, layer_name, mode='test', bins=100):
    """
    Make plots of activations of one layer of a model.

    Arrays will be flattend before plotting them as histograms.

    Parameters
    ----------
    model : keras model
        The model to make the data with.
    samples : dict
        Input data.
    layer_name : str or None
        Name of the layer to get info from. None for all layers.
    mode : str
        Mode of the layers during the forward pass. Either train or test.
        Important for batchnorm, dropout, ...
    bins : int
        Number of bins of the histogram.

    """

    layer = model.get_layer(layer_name)
    if layer.name in model.input_names:
        activations = samples[layer.name]
    else:
        activations = get_layer_output(model, samples, layer.name, mode)

    plt.hist(activations.flatten(), bins=bins)
    plt.title('Activations for layer ' + str(layer_name))
    plt.xlabel('Activation (layer output)')
    plt.ylabel('Quantity [#]')


def plot_weights(model, layer_name, bins=100):
    """
    Make plots of the weights of one layer of a model.

    Arrays will be flattend before plotting them as histograms.

    Parameters
    ----------
    model : keras model
        The model to make the data with.
    layer_name : str or None
        Name of the layer to get info from. None for all layers.
    bins : int
        Number of bins of the histogram.

    Raises
    ------
    ValueError
        If there are no weights in the given layer.

    """
    layer = model.get_layer(layer_name)
    layer_weights = layer.get_weights()

    if not layer_weights:
        raise ValueError("Layer contains no weights")

    # layer_weights is a list of np arrays; flatten it
    weights = np.array([])
    for j, w_temp in enumerate(layer_weights):
        w_temp_flattened = np.array(w_temp.flatten(), dtype=np.float64)
        weights = np.concatenate((weights, w_temp_flattened), axis=0)

    plt.hist(weights, bins=bins)
    plt.title('Weights for layer ' + str(layer_name))
    plt.xlabel('Weight')
    plt.ylabel('Quantity [#]')
    plt.tight_layout()




from unittest import TestCase
from unittest.mock import MagicMock, patch
import os
import shutil
from keras.models import Model
import keras.layers as layers

from orcanet.core import Organizer, Configuration


class TestOrganizer(TestCase):
    """
    Test the organizer by creating a dummy directory .temp/core.
    """
    def setUp(self):
        self.temp_dir = os.path.join(os.path.dirname(__file__), ".temp", "core")
        os.mkdir(self.temp_dir)

        self.orga = Organizer(self.temp_dir)

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    @patch('orcanet.core.ks.utils.plot_model')
    def test_load_model_new_training(self, mock_plot_model):
        def plot_history(model, folder):
            pass
        mock_plot_model.side_effect = plot_history

        # latest epoch is None aka no model has been trained
        latest_epoch = None
        self.orga.io.get_latest_epoch = MagicMock(return_value=latest_epoch)

        # no model given = error
        target_model = None
        with self.assertRaises(ValueError):
            self.orga._get_model(target_model, logging=False)

        # model given = ok
        target_model = build_test_model()
        model = self.orga._get_model(target_model, logging=False)
        self.assertEqual(target_model.to_json(), model.to_json())

    def test_load_model_continue_training(self):
        # latest epoch is not None aka model has been trained before
        latest_epoch = (1, 1)
        self.orga.io.get_latest_epoch = MagicMock(return_value=latest_epoch)

        # no model given = load model
        target_model = None
        model_file = os.path.join(self.temp_dir, "test_model.h5")
        self.orga.io.get_model_path = MagicMock(return_value=model_file)

        # no model given = ok
        saved_model = build_test_model()
        saved_model.save(model_file)
        loaded_model = self.orga._get_model(model=target_model, logging=False)
        self.assertEqual(saved_model.to_json(), loaded_model.to_json())

        # model given = ok
        target_model = 1.2

        model = self.orga._get_model(target_model, logging=False)
        self.assertEqual(target_model, model)

    def test_val_is_due(self):
        no_of_files = 3
        self.orga.cfg.validate_interval = 2
        targets = {
            (1, 1): False,
            (1, 2): True,
            (1, 3): True,

            (2, 1): False,
            (2, 2): True,
        }

        self.orga.io.get_no_of_files = MagicMock(return_value=no_of_files)
        for epoch, target in targets.items():
            value = self.orga.val_is_due(epoch)
            self.assertEqual(value, target)

    def test_check_if_pred_already_done(self):
        epoch, fileno = 1, 1  # these dont matter
        # latest_prediction_file_no = None, total_files = 5 -> pred_done = False
        self.orga.io.get_latest_prediction_file_no = MagicMock(return_value=None)
        self.orga.io.get_no_of_files = MagicMock(return_value=5)

        value = self.orga._check_if_pred_already_done(epoch, fileno)
        target = False
        self.assertEqual(value, target)

        # latest_prediction_file_no = 2, total_files = 5 -> pred_done = False
        self.orga.io.get_latest_prediction_file_no = MagicMock(return_value=2)
        self.orga.io.get_no_of_files = MagicMock(return_value=5)

        value = self.orga._check_if_pred_already_done(epoch, fileno)
        target = False
        self.assertEqual(value, target)

        # latest_prediction_file_no = 4, total_files = 5 -> pred_done = True
        self.orga.io.get_latest_prediction_file_no = MagicMock(return_value=2)
        self.orga.io.get_no_of_files = MagicMock(return_value=5)

        value = self.orga._check_if_pred_already_done(epoch, fileno)
        target = False
        self.assertEqual(value, target)

    # @patch('orcanet.core.make_model_prediction')  TODO with all the mocks this doesnt make sense anymore.
    # def test_predict(self, mocked):
    #     self.orga.cfg._list_file = "test/test.toml"
    #     func_args = []
    #
    #     def mock_make_model_prediction(orga, model, epoch, fileno, latest_val_file_no, samples=None):
    #         func_args.extend([orga, model, epoch, fileno, latest_val_file_no, samples])
    #
    #     mocked.side_effect = mock_make_model_prediction
    #
    #     self.orga.load_saved_model = MagicMock(return_value=build_test_model())
    #     self.orga.io.get_latest_prediction_file_no = MagicMock(return_value=1)
    #     self.orga.io.get_no_of_files = MagicMock(return_value=1)
    #     self.orga.io.get_pred_files_list = MagicMock(return_value=
    #                                                    ['pred_model_epoch_2_file_2_on_bg_list_val_file_0.h5',
    #                                                     'pred_model_epoch_2_file_2_on_bg_list_val_file_1.h5'])
    #
    #     value_filepath = self.orga.predict(epoch=1, fileno=1)
    #     target_filepath = "./predictions/pred_model_epoch_1_file_1_on_test_val_files.h5"
    #
    #     self.assertEqual(value_filepath, target_filepath)
    #     self.assertEqual(func_args[2], value_filepath)
    #     self.assertEqual(func_args[3], None)


class TestZeroCenter(TestCase):
    def setUp(self):
        self.orga = Organizer("./")
        self.orga.cfg.zero_center_folder = "nope/"

    def test_get_xs_mean_existing(self):
        target = "asd"
        self.orga.xs_mean = target
        value = self.orga.get_xs_mean(logging=False)
        self.assertEqual(target, value)

    def test_get_xs_mean_no_folder(self):
        self.orga.cfg.zero_center_folder = None
        with self.assertRaises(ValueError):
            self.orga.get_xs_mean(logging=False)

    @patch('orcanet.core.load_zero_center_data')
    def test_get_xs_mean_new(self, load_mock):
        target = "blub"

        def mock_xs_mean():
            return target

        load_mock.return_value = mock_xs_mean()
        xs_mean = self.orga.get_xs_mean(logging=False)
        self.assertEqual(target, xs_mean)
        self.assertEqual(target, self.orga.xs_mean)


class TestConfiguration(TestCase):
    def setUp(self):
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.working_list = self.data_folder + "/in_out_test_list.toml"
        self.temp_dir = os.path.join(os.path.dirname(__file__), ".temp", "core")
        os.mkdir(self.temp_dir)

    def tearDown(self):
        os.rmdir(self.temp_dir)

    def test_output_folder_bracket(self):
        list_file = None
        config_file = None

        output_folder = "test"
        cfg = Configuration(output_folder, list_file, config_file)
        self.assertEqual(output_folder + "/", cfg.output_folder)

        output_folder = "test/"
        cfg = Configuration(output_folder, list_file, config_file)
        self.assertEqual(output_folder, cfg.output_folder)

    def test_listfile_already_given(self):
        config_file = None
        output_folder = "test"

        list_file = self.working_list
        cfg = Configuration(output_folder, list_file, config_file)

        with self.assertRaises(ValueError):
            cfg.import_list_file(list_file)

    def test_get_files_not_given(self):
        cfg = Configuration("./", None, None)
        with self.assertRaises(AttributeError):
            cfg.get_files("train")
        with self.assertRaises(AttributeError):
            cfg.get_files("val")
        with self.assertRaises(AttributeError):
            cfg.get_files("inference")
        with self.assertRaises(NameError):
            cfg.get_files("schmu")

    def test_update_config_bad_format(self):
        cfg = Configuration("./", None, None)

        content = [
            "[config]",
            "schmu=1",
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(AttributeError):
                cfg.update_config(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_ok(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',
            'inference_files = ["input_A_inf_1.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation_files = ["input_B_val_1.h5",]',
            'inference_files = ["input_B_inf_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            cfg.import_list_file(file_name)

        finally:
            os.remove(file_name)

    def test_load_list_file_unknown_keyword(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',
            'schmu = 1',
            
            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(NameError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_no_train_files(self):
        cfg = Configuration("./", None, None)
        val_files = {"input_A": ("input_A_val_1.h5", ),
                     "input_B": ("input_B_val_1.h5", )}
        content = [
            '[input_A]',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            cfg.import_list_file(file_name)
            with self.assertRaises(AttributeError):
                cfg.get_files("train")
            with self.assertRaises(AttributeError):
                cfg.get_files("inference")

            self.assertEqual(val_files, cfg.get_files("val"))

        finally:
            os.remove(file_name)

    def test_load_list_file_no_val_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'schmu=1',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(NameError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_different_no_of_train_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", ]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(ValueError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_different_no_of_val_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5", "input_A_val_2.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")

            with self.assertRaises(ValueError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)


def build_test_model(compile=False):
    input_shape = (1,)
    inp = layers.Input(input_shape, name="inp")
    outp = layers.Dense(1, name="out")(inp)

    test_model = Model(inp, outp)
    if compile:
        test_model.compile("sgd", loss={"out": "mse"}, metrics={"out": "mae"})

    return test_model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code for training and validating NN's, as well as evaluating them.
"""

import h5py
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from orcanet.utilities.layer_plotting import plot_activations, plot_weights
from orcanet.logging import BatchLogger
from orcanet.in_out import h5_get_number_of_rows
from orcanet.h5_generator import get_h5_generator

# for debugging
# from tensorflow.python import debug as tf_debug
# K.set_session(tf_debug.LocalCLIDebugWrapperSession(tf.Session()))


def train_model(orga, model, epoch, batch_logger=False):
    """
    Train a model on one file and return the history.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : keras model
        A compiled keras model.
    epoch : tuple
        Current epoch and the no of the file to train on.
    batch_logger : bool
        Use the orcanet batchlogger to log the training.

    Returns
    -------
    history : dict
        The history of the training on this file. A record of training
        loss values and metrics values.

    """
    files_dict = orga.io.get_file("train", epoch[1])

    if orga.cfg.n_events is not None:
        # TODO Can throw an error if n_events is larger than the file
        f_size = orga.cfg.n_events  # for testing purposes
    else:
        f_size = orga.io.get_file_sizes("train")[epoch[1] - 1]

    callbacks = []
    if batch_logger:
        callbacks.append(BatchLogger(orga, epoch))
    if orga.cfg.callback_train is not None:
        try:
            callbacks.extend(orga.cfg.callback_train)
        except TypeError:
            callbacks.append(orga.cfg.callback_train)

    training_generator = get_h5_generator(
        orga, files_dict, f_size=f_size,
        zero_center=orga.cfg.zero_center_folder is not None,
        shuffle=orga.cfg.shuffle_train)

    history = model.fit_generator(
        training_generator,
        steps_per_epoch=int(f_size / orga.cfg.batchsize),
        verbose=orga.cfg.verbose_train,
        max_queue_size=orga.cfg.max_queue_size,
        callbacks=callbacks,
        initial_epoch=epoch[0] - 1,
        epochs=epoch[0],
    )

    # get a dict with losses and metrics
    # only trained for one epoch, so value is list of len 1
    history = {key: value[0] for key, value in history.history.items()}
    return history


def validate_model(orga, model):
    """
    Validates a model on all validation files and return the history.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : keras model
        A compiled keras model.

    Returns
    -------
    history : dict
        The history of the validation on all files. A record of validation
        loss values and metrics values.

    """
    # One history for each val file
    histories = []
    f_sizes = orga.io.get_file_sizes("val")

    for i, files_dict in enumerate(orga.io.yield_files("val")):
        f_size = f_sizes[i]
        if orga.cfg.n_events is not None:
            f_size = orga.cfg.n_events  # for testing purposes

        val_generator = get_h5_generator(
            orga, files_dict, f_size=f_size,
            zero_center=orga.cfg.zero_center_folder is not None)

        history_file = model.evaluate_generator(
            val_generator,
            steps=int(f_size / orga.cfg.batchsize),
            max_queue_size=orga.cfg.max_queue_size,
            verbose=orga.cfg.verbose_val)
        if not isinstance(history_file, list):
            history_file = [history_file, ]
        histories.append(history_file)

    # average over all val files
    history = weighted_average(histories, f_sizes)

    # This history is just a list, not a dict like with fit_generator
    # so transform to dict
    history = dict(zip(model.metrics_names, history))

    return history


def weighted_average(histories, f_sizes):
    """
    Average multiple histories, weighted with the file size.

    Each history can have multiple metrics, which are averaged seperatly.

    Parameters
    ----------
    histories : List
        List of histories, one for each file. Each history is also
        a list: each entry is a different loss or metric.
    f_sizes : List
        List of the file sizes, in the same order as the histories, i.e.
        the file of histories[0] has the length f_sizes[0].

    Returns
    -------
    wgtd_average : List
        The weighted averaged history. Has the same length as each
        history in the histories List, i.e. one entry per loss or metric.

    """
    assert len(histories) == len(f_sizes)
    rltv_fsizes = [f_size/sum(f_sizes) for f_size in f_sizes]
    wgtd_average = np.dot(np.transpose(histories), rltv_fsizes)

    return wgtd_average.tolist()


def save_actv_wghts_plot(orga, model, epoch, samples=1):
    """
    Plots the weights of a model and the activations for samples from
    the validation set to one .pdf file each.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : ks.models.Model
        The model to do the predictions with.
    epoch : tuple
        Current epoch and fileno.
    samples : int
        Number of samples to make the plot for.

    """
    plt.ioff()

    file = next(orga.io.yield_files("val"))
    generator = get_h5_generator(
        orga, file, f_size=samples,
        zero_center=orga.cfg.zero_center_folder is not None,
        keras_mode=True)
    xs, ys = next(generator)

    pdf_name_act = "{}/activations_epoch_{}_file_{}.pdf".format(
        orga.io.get_subfolder("activations", create=True), epoch[0], epoch[1])

    with PdfPages(pdf_name_act) as pdf:
        for layer in model.layers:
            plot_activations(model, xs, layer.name, mode='test')
            pdf.savefig()
            plt.clf()
        plt.close()

    pdf_name_wght = "{}/weights_epoch_{}_file_{}.pdf".format(
        orga.io.get_subfolder("activations", create=True), epoch[0], epoch[1])

    with PdfPages(pdf_name_wght) as pdf:
        for layer in model.layers:
            try:
                plot_weights(model, layer.name)
            except ValueError:
                continue
            pdf.savefig()
            plt.clf()
        plt.close()


def h5_inference(orga, model, files_dict, output_path, samples=None, use_def_label=True):
    """
    Let a model predict on all samples in a h5 file, and save it as a h5 file.

    Per default, the h5 file will contain a datagroup y_values straight from
    the given files, as well as two datagroups per output layer of the network,
    which have the labels and the predicted values in them as numpy arrays,
    respectively.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : ks.model.Model
        Trained Keras model of a neural network.
    files_dict : dict
        Dict mapping model input names to h5 file paths.
    output_path : str
        Name of the output h5 file containing the predictions.
    samples : int, optional
        Dont use all events in the file, but instead only the given number.
    use_def_label : bool
        If True and no label modifier is given by user, use the default
        label modifier instead of none.

    """
    batchsize = orga.cfg.batchsize
    compression = ("gzip", 1)

    file_size = h5_get_number_of_rows(
        list(files_dict.values())[0],
        datasets=[orga.cfg.key_x_values])
    generator = get_h5_generator(
        orga,
        files_dict,
        zero_center=orga.cfg.zero_center_folder is not None,
        keras_mode=False,
        use_def_label=use_def_label)

    if samples is None:
        steps = int(file_size / batchsize)
        if file_size % batchsize != 0:
            # add a smaller step in the end
            steps += 1
    else:
        steps = int(samples / batchsize)

    with h5py.File(output_path, 'w') as h5_file:
        for s in range(steps):
            if s % 1000 == 0:
                print('Predicting in step {}/{} ({:0.2%})'.format(
                    s, steps, s/steps))

            info_blob = next(generator)

            y_pred = model.predict_on_batch(info_blob["xs"])
            if not isinstance(y_pred, list):
                # if only one output, transform to a list
                y_pred = [y_pred]
            # transform y_pred to dict
            y_pred = {
                out: y_pred[i] for i, out in enumerate(model.output_names)}
            info_blob["y_pred"] = y_pred

            if orga.cfg.dataset_modifier is None:
                datasets = get_datasets(info_blob)
            else:
                datasets = orga.cfg.dataset_modifier(info_blob)

            # TODO maybe add attr to data, like used files or orcanet version number?
            if s == 0:  # create datasets in the first step
                for dataset_name, data in datasets.items():
                    maxshape = (file_size,) + data.shape[1:]
                    chunks = True  # (batchsize,) + data.shape[1:]
                    h5_file.create_dataset(
                        dataset_name, data=data, maxshape=maxshape,
                        chunks=chunks, compression=compression[0],
                        compression_opts=compression[1])

            else:
                for dataset_name, data in datasets.items():
                    # append data at the end of the dataset
                    h5_file[dataset_name].resize(
                        h5_file[dataset_name].shape[0] + data.shape[0], axis=0)
                    h5_file[dataset_name][-data.shape[0]:] = data


def make_model_prediction(orga, model, epoch, fileno, samples=None):
    """
    Let a model predict on all validation samples, and save it as a h5 file.

    Per default, the h5 file will contain a datagroup y_values straight from
    the given files, as well as two datagroups per output layer of the network,
    which have the labels and the predicted values in them as numpy arrays,
    respectively.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : ks.model.Model
        Trained Keras model of a neural network.
    epoch : int
        Epoch of the last model training step in the epoch, file_no tuple.
    fileno : int
        File number of the last model training step in the epoch, file_no tuple.
    samples : int or None
        Number of events that should be predicted.
        If samples=None, the whole file will be used.

    """
    latest_pred_file_no = orga.io.get_latest_prediction_file_no(epoch, fileno)
    if latest_pred_file_no is None:
        latest_pred_file_no = 0

    # For every val file set (one set can have multiple files if
    # the model has multiple inputs):
    for f_number, files_dict in enumerate(orga.io.yield_files("val"), 1):
        if f_number <= latest_pred_file_no:
            continue

        pred_filepath = orga.io.get_pred_path(epoch, fileno, f_number)
        h5_inference(orga, model, files_dict, pred_filepath, samples=samples)


def get_datasets(info_blob):
    """
    Get the dataset names and numpy array contents.

    Every output layer will get one dataset each for both the label and
    the prediction. E.g. if your model has an output layer called "energy",
    the datasets "label_energy" and "pred_energy" will be made.
    If there are no labels (e.g. during orga.inference), the label dataset
    will not be generated.

    Parameters
    ----------
    info_blob : dict
        Contains the following infos as keys/values:
        xs : ndarray
            Input to the network, after sample modifier has been applied.
        y_values : ndarray or None
            A structured array containing infos for every event, right from
            the input files.
            Can also be None, e.g. for when there are no y_values.
        ys : dict or None
            The labels for each output layer of the network.
            Can also be None, e.g. for when there are no y_values.
        y_pred : dict
            The predictions of each output layer of the network.

    Returns
    -------
    datasets : dict
        Keys are the name of the datagroups, values the content in the
        form of numpy arrays.

    """
    datasets = dict()
    if "y_values" in info_blob:
        datasets["y_values"] = info_blob["y_values"]

    if info_blob.get("ys") is not None:
        y_true = info_blob["ys"]
        for out_layer_name in y_true:
            datasets["label_" + out_layer_name] = y_true[out_layer_name]

    y_pred = info_blob["y_pred"]
    for out_layer_name in y_pred:
        datasets["pred_" + out_layer_name] = y_pred[out_layer_name]

    return datasets



from unittest import TestCase
from orcanet.core import Organizer
import orcanet.model_builder
from orcanet_contrib.custom_objects import get_custom_objects
import os


class TestSequentialBuilder(TestCase):
    """
    Just build some models to see if there are embarrassing formatting errors.
    """
    def setUp(self):
        orca_dir = os.path.dirname(os.path.dirname(orcanet.model_builder.__file__))
        self.example_dir = os.path.join(orca_dir, "examples", "model_files")

        def get_orga(dims):
            orga = Organizer(".")
            if dims == 2:
                orga.io.get_input_shapes = get_input_shapes_2d
            elif dims == 3:
                orga.io.get_input_shapes = get_input_shapes_3d
            else:
                raise AssertionError
            return orga
        self.get_orga = get_orga

    def test_cnn(self):
        toml_file = "cnn.toml"

        model_file = os.path.join(self.example_dir, toml_file)
        mb = orcanet.model_builder.ModelBuilder(model_file)
        model = mb.build(self.get_orga(dims=2))

    def test_resnet(self):
        toml_file = "resnet.toml"

        model_file = os.path.join(self.example_dir, toml_file)
        mb = orcanet.model_builder.ModelBuilder(model_file)
        model = mb.build(self.get_orga(dims=2))

    def test_explanation(self):
        toml_file = "explanation.toml"

        model_file = os.path.join(self.example_dir, toml_file)
        mb = orcanet.model_builder.ModelBuilder(model_file)
        orga = self.get_orga(dims=3)
        orga.cfg.custom_objects = get_custom_objects()
        model = mb.build(orga)


def get_input_shapes_3d():
    dims = (10, 10, 10, 1)
    return {"input_A": dims}


def get_input_shapes_2d():
    dims = (10, 10, 1)
    return {"input_A": dims}



import keras as ks
import keras.layers as layers


class ConvBlock:
    def __init__(self, conv_dim,
                 filters,
                 kernel_size=3,
                 strides=1,
                 pool_size=None,
                 pool_padding="valid",
                 dropout=None,
                 sdropout=None,
                 activation='relu',
                 kernel_l2_reg=None,
                 batchnorm=False,
                 kernel_initializer="he_normal"):
        """
        2D/3D Convolutional block followed by BatchNorm, Activation,
        MaxPooling and/or Dropout.

        Parameters
        ----------
        conv_dim : int
            Specifies the dimension of the convolutional block, 2D/3D.
        filters : int
            Number of filters used for the convolutional layer.
        strides : int or tuple
            The stride length of the convolution
        kernel_size : int or tuple
            Kernel size which is used for all three dimensions.
        pool_size : None or tuple
            Specifies if a MaxPooling layer should be added. e.g. (1,1,2)
            -> sizes for a 3D conv block.
        dropout : float or None
            Adds a dropout layer if the value is not None.
            Can not be used together with sdropout.
        sdropout : float or None
            Adds a spatial dropout layer if the value is not None.
            Can not be used together with dropout.
        activation : str or None
            Type of activation function that should be used. E.g. 'linear',
            'relu', 'elu', 'selu'.
        kernel_l2_reg : float, optional
            Regularization factor of l2 regularizer for the weights.
        batchnorm : bool
            Adds a batch normalization layer.
        kernel_initializer : string
            Initializer for the kernel weights.

        """
        self.conv_dim = conv_dim
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.pool_size = pool_size
        self.pool_padding = pool_padding
        self.dropout = dropout
        self.sdropout = sdropout
        self.activation = activation
        self.kernel_l2_reg = kernel_l2_reg
        self.batchnorm = batchnorm
        self.kernel_initializer = kernel_initializer

    def __call__(self, inputs):
        if self.dropout is not None and self.sdropout is not None:
            raise ValueError("Can only use either dropout or spatial "
                             "dropout, not both")
        if self.conv_dim == 2:
            convolution_nd = layers.Convolution2D
            max_pooling_nd = layers.MaxPooling2D
            s_dropout_nd = layers.SpatialDropout2D
        elif self.conv_dim == 3:
            convolution_nd = layers.Convolution3D
            max_pooling_nd = layers.MaxPooling3D
            s_dropout_nd = layers.SpatialDropout3D
        else:
            raise ValueError('dim must be equal to 2 or 3.')

        if self.kernel_l2_reg is not None:
            kernel_reg = ks.regularizers.l2(self.kernel_l2_reg)
        else:
            kernel_reg = None

        if self.batchnorm:
            use_bias = False
        else:
            use_bias = True

        x = convolution_nd(filters=self.filters,
                           kernel_size=self.kernel_size,
                           strides=self.strides,
                           padding='same',
                           kernel_initializer=self.kernel_initializer,
                           use_bias=use_bias,
                           kernel_regularizer=kernel_reg)(inputs)

        if self.batchnorm:
            channel_axis = 1 if ks.backend.image_data_format() == "channels_first" else -1
            x = layers.BatchNormalization(axis=channel_axis)(x)
        if self.activation is not None:
            x = layers.Activation(self.activation)(x)
        if self.pool_size is not None:
            x = max_pooling_nd(pool_size=self.pool_size,
                               padding=self.pool_padding)(x)
        if self.dropout is not None:
            x = layers.Dropout(self.dropout)(x)
        elif self.sdropout is not None:
            x = s_dropout_nd(self.sdropout)(x)

        return x


class DenseBlock:
    def __init__(self, units,
                 dropout=None,
                 activation='relu',
                 kernel_l2_reg=None,
                 batchnorm=False,
                 kernel_initializer="he_normal"):
        """
        Dense layer followed by BatchNorm, Activation and/or Dropout.

        Parameters
        ----------
        units : int
            Number of neurons of the dense layer.
        dropout : float or None
            Adds a dropout layer if the value is not None.
        activation : str or None
            Type of activation function that should be used. E.g. 'linear',
            'relu', 'elu', 'selu'.
        kernel_l2_reg : float, optional
            Regularization factor of l2 regularizer for the weights.
        batchnorm : bool
            Adds a batch normalization layer.

        """
        self.units = units
        self.dropout = dropout
        self.activation = activation
        self.kernel_l2_reg = kernel_l2_reg
        self.batchnorm = batchnorm
        self.kernel_initializer = kernel_initializer

    def __call__(self, inputs):
        if self.kernel_l2_reg is not None:
            kernel_reg = ks.regularizers.l2(self.kernel_l2_reg)
        else:
            kernel_reg = None

        if self.batchnorm:
            use_bias = False
        else:
            use_bias = True

        x = layers.Dense(
            units=self.units,
            use_bias=use_bias,
            kernel_initializer=self.kernel_initializer,
            kernel_regularizer=kernel_reg)(inputs)

        if self.batchnorm:
            channel_axis = 1 if ks.backend.image_data_format() == "channels_first" else -1
            x = layers.BatchNormalization(axis=channel_axis)(x)
        if self.activation is not None:
            x = layers.Activation(self.activation)(x)
        if self.dropout is not None:
            x = layers.Dropout(self.dropout)(x)
        return x


class ResnetBlock:
    def __init__(self, conv_dim,
                 filters,
                 strides=1,
                 kernel_size=3,
                 activation='relu',
                 batchnorm=False,
                 kernel_initializer="he_normal"):
        """
        A residual building block for resnets. 2 c layers with a shortcut.
        https://arxiv.org/pdf/1605.07146.pdf

        Parameters
        ----------
        conv_dim : int
            Specifies the dimension of the convolutional block, 2D/3D.
        filters : int
            Number of filters used for the convolutional layers.
        strides : int or tuple
            The stride length of the convolution. If strides is 1, this is
            the identity block. If not, it has a conv block
            at the shortcut.
        kernel_size : int or tuple
            Kernel size which is used for all three dimensions.
        activation : str or None
            Type of activation function that should be used. E.g. 'linear',
            'relu', 'elu', 'selu'.
        batchnorm : bool
            Adds a batch normalization layer.
        kernel_initializer : string
            Initializer for the kernel weights.

        """
        self.conv_dim = conv_dim
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.activation = activation
        self.batchnorm = batchnorm
        self.kernel_initializer = kernel_initializer

    def __call__(self, inputs):
        x = ConvBlock(conv_dim=self.conv_dim,
                      filters=self.filters,
                      kernel_size=self.kernel_size,
                      strides=self.strides,
                      kernel_initializer=self.kernel_initializer,
                      batchnorm=self.batchnorm,
                      activation=self.activation)(inputs)
        x = ConvBlock(conv_dim=self.conv_dim,
                      filters=self.filters,
                      kernel_size=self.kernel_size,
                      kernel_initializer=self.kernel_initializer,
                      batchnorm=self.batchnorm,
                      activation=None)(x)

        if self.strides != 1:
            shortcut = ConvBlock(conv_dim=self.conv_dim,
                                 filters=self.filters,
                                 kernel_size=1,
                                 strides=self.strides,
                                 kernel_initializer=self.kernel_initializer,
                                 activation=None,
                                 batchnorm=self.batchnorm,
                                 )(inputs)
        else:
            shortcut = inputs

        x = layers.add([x, shortcut])
        x = layers.Activation(self.activation)(x)
        return x


class ResnetBnetBlock:
    def __init__(self, conv_dim,
                 filters,
                 strides=1,
                 kernel_size=3,
                 activation='relu',
                 batchnorm=False,
                 kernel_initializer="he_normal"):
        """
        A residual bottleneck building block for resnets.
        https://arxiv.org/pdf/1605.07146.pdf

        Parameters
        ----------
        conv_dim : int
            Specifies the dimension of the convolutional block, 2D/3D.
        filters : List
            Number of filters used for the convolutional layers.
            Has to be length 3. First and third is for the 1x1 convolutions.
        strides : int or tuple
            The stride length of the convolution. If strides is 1, this is
            the identity block. If not, it has a conv block
            at the shortcut.
        kernel_size : int or tuple
            Kernel size which is used for all three dimensions.
        activation : str or None
            Type of activation function that should be used. E.g. 'linear',
            'relu', 'elu', 'selu'.
        batchnorm : bool
            Adds a batch normalization layer.
        kernel_initializer : string
            Initializer for the kernel weights.

        """
        self.conv_dim = conv_dim
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.activation = activation
        self.batchnorm = batchnorm
        self.kernel_initializer = kernel_initializer

    def __call__(self, inputs):
        filters1, filters2, filters3 = self.filters

        x = ConvBlock(conv_dim=self.conv_dim,
                      filters=filters1,
                      kernel_size=1,
                      strides=self.strides,
                      kernel_initializer=self.kernel_initializer,
                      batchnorm=self.batchnorm,
                      activation=self.activation)(inputs)
        x = ConvBlock(conv_dim=self.conv_dim,
                      filters=filters2,
                      kernel_size=self.kernel_size,
                      kernel_initializer=self.kernel_initializer,
                      batchnorm=self.batchnorm,
                      activation=self.activation)(x)
        x = ConvBlock(conv_dim=self.conv_dim,
                      filters=filters3,
                      kernel_size=1,
                      kernel_initializer=self.kernel_initializer,
                      batchnorm=self.batchnorm,
                      activation=None)(x)

        if self.strides != 1:
            shortcut = ConvBlock(conv_dim=self.conv_dim,
                                 filters=filters3,
                                 kernel_size=1,
                                 strides=self.strides,
                                 kernel_initializer=self.kernel_initializer,
                                 activation=None,
                                 batchnorm=self.batchnorm,
                                 )(inputs)
        else:
            shortcut = inputs

        x = layers.add([x, shortcut])
        x = layers.Activation(self.activation)(x)
        return x


def input_block(input_shapes):
    """
    Build input layers according to a dict mapping the layer names to shapes.

    Parameters
    ----------
    input_shapes : dict
        Keys: Input layer names.
        Values: Their shapes.

    Returns
    -------
    inputs : List
        A list of named keras input layers.

    """
    inputs = []
    for input_name, input_shape in input_shapes.items():
        inputs.append(layers.Input(
            shape=input_shape, name=input_name, dtype=ks.backend.floatx()))
    return inputs
